@@ -34,6 +34,8 @@
 #include "files.h"
 #include "sorting.h"
 
+/*+ Enable debugging print statements +*/
+#define DEBUG 0
 
 /* Global variables */
 
@@ -71,19 +73,27 @@ typedef struct _thread_data
  }
  thread_data;
 
-/* Thread variables */
+/* Thread variables and functions */
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
 static pthread_mutex_t running_mutex = PTHREAD_MUTEX_INITIALIZER;
 static pthread_cond_t  running_cond  = PTHREAD_COND_INITIALIZER;
 
+static int wait_for_free_thread(thread_data *threads,int nthreads,int *threads_running);
+static void wait_for_all_threads(thread_data *threads,int nthreads,int *threads_running);
+
 #endif
 
 /* Thread helper functions */
 
-static void *filesort_fixed_heapsort_thread(thread_data *thread);
-static void *filesort_vary_heapsort_thread(thread_data *thread);
+static void *filesort_heapsort_thread(thread_data *thread);
+
+/* Local functions */
+
+static index_t filesort_merge(int fd_out,int nfiles,size_t itemsize,size_t largestitemsize,
+                              int(*compare_function)(const void*, const void*),
+                              int (*post_sort_function)(void*,index_t));
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -122,17 +132,14 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
                                                             int (*compare_function)(const void*,const void*),
                                                             int (*post_sort_function)(void*,index_t))
 {
- int *fds=NULL,*heap=NULL;
- int nfiles=0,ndata=0;
+ int nfiles=0;
  index_t count_out=0,count_in=0,total=0;
  size_t nitems,item;
- char *data;
- void **datap;
  thread_data *threads;
- int i,more=1;
+ int nthreads,i,more=1;
  char *filename=(char*)malloc_logassert(strlen(option_tmpdirname)+24);
 #if defined(USE_PTHREADS) && USE_PTHREADS
- int nthreads=0;
+ int threads_running=0;
 #endif
 
  /* Allocate the RAM buffer and other bits */
@@ -142,16 +149,26 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
  if(nitems==0)
     return(0);
 
- if((nitems*(itemsize+sizeof(void*)))<(option_filesort_ramsize/option_filesort_threads))
-    /* use one thread */;
+ if(option_filesort_threads==1)
+    nthreads = 1;
  else if((nitems*(itemsize+sizeof(void*)))<option_filesort_ramsize)
-    nitems=1+nitems/option_filesort_threads;
+    nthreads = 1 + (nitems*(itemsize+sizeof(void*))) / (option_filesort_ramsize/option_filesort_threads);
  else
-    nitems=option_filesort_ramsize/(option_filesort_threads*(itemsize+sizeof(void*)));
+    nthreads = option_filesort_threads;
 
- threads=(thread_data*)calloc_logassert(option_filesort_threads,sizeof(thread_data));
+ if((nitems*(itemsize+sizeof(void*)))>option_filesort_ramsize)
+    nitems=option_filesort_ramsize/(itemsize+sizeof(void*));
 
- for(i=0;i<option_filesort_threads;i++)
+ nitems=1+nitems/nthreads;
+
+#if DEBUG
+ printf("DEBUG: filesort_fixed nitems=%lu option_ramsize/option_threads=%lu => nitems=%lu datasize=%lu nthreads=%d\n",
+        (size_t)SizeFileFD(fd_in)/itemsize,option_filesort_ramsize/option_filesort_threads,nitems,nitems*(itemsize+sizeof(void*)),nthreads);
+#endif
+
+ threads=(thread_data*)calloc_logassert(nthreads,sizeof(thread_data));
+
+ for(i=0;i<nthreads;i++)
    {
     threads[i].fd=-1;
 
@@ -163,6 +180,10 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
 
     threads[i].itemsize=itemsize;
     threads[i].compare=compare_function;
+
+#if defined(USE_PTHREADS) && USE_PTHREADS
+    threads[i].running=0;
+#endif
    }
 
  /* Loop around, fill the buffer, sort the data and write a temporary file */
@@ -173,42 +194,10 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
-    if(option_filesort_threads>1)
-      {
-       /* If all threads are in use wait for an existing thread to finish */
+    /* Choose a thread to use */
 
-       if(nthreads==option_filesort_threads)
-         {
-          pthread_mutex_lock(&running_mutex);
-
-          while(nthreads==option_filesort_threads)
-            {
-             for(i=0;i<option_filesort_threads;i++)
-                if(threads[i].running==2)
-                  {
-                   pthread_join(threads[i].thread,NULL);
-                   threads[i].running=0;
-                   CloseFileBuffered(threads[i].fd);
-                   nthreads--;
-                  }
-
-             if(nthreads==option_filesort_threads)
-                pthread_cond_wait(&running_cond,&running_mutex);
-            }
-
-          pthread_mutex_unlock(&running_mutex);
-         }
-
-       /* Find a spare slot */
-
-       pthread_mutex_lock(&running_mutex);
-
-       for(thread=0;thread<option_filesort_threads;thread++)
-          if(!threads[thread].running)
-             break;
-
-       pthread_mutex_unlock(&running_mutex);
-      }
+    if(nthreads>1)
+       thread=wait_for_free_thread(threads,nthreads,&threads_running);
 
 #endif
 
@@ -233,7 +222,9 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
        count_in++;
       }
 
-    threads[thread].n=item;
+#if DEBUG
+    printf("DEBUG: filesort_fixed thread=%d file=%d item=%lu total=%u more=%d\n",thread,nfiles,item,total,more);
+#endif
 
     /* Shortcut if there is no previous data and no more data (i.e. no data at all) */
 
@@ -242,14 +233,21 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
 
     /* No new data read in this time round */
 
-    if(threads[thread].n==0)
+    if(item==0)
        break;
 
+    /* Update the number of items to sort */
+
+    threads[thread].n=item;
+
     /* Shortcut if only one file, don't write to disk */
 
     if(more==0 && nfiles==0)
-       filesort_heapsort(threads[thread].datap,threads[thread].n,threads[thread].compare);
+      {
+       filename[0]=0;
 
+       filesort_heapsort_thread(&threads[thread]);
+      }
     else
       {
        /* Create the file descriptor (not thread-safe) */
@@ -258,9 +256,9 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
 
        threads[thread].fd=OpenFileBufferedNew(filename);
 
-       if(option_filesort_threads==1)
+       if(nthreads==1)
          {
-          filesort_fixed_heapsort_thread(&threads[thread]);
+          filesort_heapsort_thread(&threads[thread]);
 
           CloseFileBuffered(threads[thread].fd);
          }
@@ -269,15 +267,11 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
 
        else
          {
-          pthread_mutex_lock(&running_mutex);
-
           threads[thread].running=1;
 
-          pthread_mutex_unlock(&running_mutex);
+          pthread_create(&threads[thread].thread,NULL,(void* (*)(void*))filesort_heapsort_thread,&threads[thread]);
 
-          pthread_create(&threads[thread].thread,NULL,(void* (*)(void*))filesort_fixed_heapsort_thread,&threads[thread]);
-
-          nthreads++;
+          threads_running++;
          }
 
 #endif
@@ -288,27 +282,12 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
    }
  while(more);
 
- /* Wait for all of the threads to finish */
-
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
- while(option_filesort_threads>1 && nthreads)
-   {
-    pthread_mutex_lock(&running_mutex);
+ /* Wait for all of the threads to finish */
 
-    pthread_cond_wait(&running_cond,&running_mutex);
-
-    for(i=0;i<option_filesort_threads;i++)
-       if(threads[i].running==2)
-         {
-          pthread_join(threads[i].thread,NULL);
-          threads[i].running=0;
-          CloseFileBuffered(threads[i].fd);
-          nthreads--;
-         }
-
-    pthread_mutex_unlock(&running_mutex);
-   }
+ if(nthreads>1)
+    wait_for_all_threads(threads,nthreads,&threads_running);
 
 #endif
 
@@ -330,145 +309,21 @@ index_t filesort_fixed(int fd_in,int fd_out,size_t itemsize,int (*pre_sort_funct
          }
       }
 
-    DeleteFile(filename);
+    if(filename[0])
+       DeleteFile(filename);
 
     goto tidy_and_exit;
    }
 
- /* Check that number of files is less than file size */
+ /* Open all of the temporary files and perform an n-way merge using a binary heap */
 
- logassert((unsigned)nfiles<nitems,"Too many temporary files (use more sorting memory?)");
-
- /* Open all of the temporary files */
-
- fds=(int*)malloc_logassert(nfiles*sizeof(int));
-
- for(i=0;i<nfiles;i++)
-   {
-    sprintf(filename,"%s/filesort.%d.tmp",option_tmpdirname,i);
-
-    fds[i]=ReOpenFileBuffered(filename);
-
-    DeleteFile(filename);
-   }
-
- /* Perform an n-way merge using a binary heap */
-
- heap=(int*)malloc_logassert((1+nfiles)*sizeof(int));
-
- data =threads[0].data;
- datap=threads[0].datap;
-
- /* Fill the heap to start with */
-
- for(i=0;i<nfiles;i++)
-   {
-    int index;
-
-    datap[i]=data+i*itemsize;
-
-    ReadFileBuffered(fds[i],datap[i],itemsize);
-
-    index=i+1;
-
-    heap[index]=i;
-
-    /* Bubble up the new value */
-
-    while(index>1)
-      {
-       int newindex;
-       int temp;
-
-       newindex=index/2;
-
-       if(compare_function(datap[heap[index]],datap[heap[newindex]])>=0)
-          break;
-
-       temp=heap[index];
-       heap[index]=heap[newindex];
-       heap[newindex]=temp;
-
-       index=newindex;
-      }
-   }
-
- /* Repeatedly pull out the root of the heap and refill from the same file */
-
- ndata=nfiles;
-
- do
-   {
-    int index=1;
-
-    if(!post_sort_function || post_sort_function(datap[heap[index]],count_out))
-      {
-       WriteFileBuffered(fd_out,datap[heap[index]],itemsize);
-       count_out++;
-      }
-
-    if(ReadFileBuffered(fds[heap[index]],datap[heap[index]],itemsize))
-      {
-       heap[index]=heap[ndata];
-       ndata--;
-      }
-
-    /* Bubble down the new value */
-
-    while((2*index)<ndata)
-      {
-       int newindex;
-       int temp;
-
-       newindex=2*index;
-
-       if(compare_function(datap[heap[newindex]],datap[heap[newindex+1]])>=0)
-          newindex=newindex+1;
-
-       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
-          break;
-
-       temp=heap[newindex];
-       heap[newindex]=heap[index];
-       heap[index]=temp;
-
-       index=newindex;
-      }
-
-    if((2*index)==ndata)
-      {
-       int newindex;
-       int temp;
-
-       newindex=2*index;
-
-       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
-          ; /* break */
-       else
-         {
-          temp=heap[newindex];
-          heap[newindex]=heap[index];
-          heap[index]=temp;
-         }
-      }
-   }
- while(ndata>0);
+ count_out=filesort_merge(fd_out,nfiles,itemsize,0,compare_function,post_sort_function);
 
  /* Tidy up */
 
  tidy_and_exit:
 
- if(fds)
-   {
-    for(i=0;i<nfiles;i++)
-       CloseFileBuffered(fds[i]);
-    free(fds);
-   }
-
- if(heap)
-    free(heap);
-
- for(i=0;i<option_filesort_threads;i++)
+ for(i=0;i<nthreads;i++)
    {
     log_free(threads[i].data);
     log_free(threads[i].datap);
@@ -519,18 +374,15 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
                                            int (*compare_function)(const void*,const void*),
                                            int (*post_sort_function)(void*,index_t))
 {
- int *fds=NULL,*heap=NULL;
- int nfiles=0,ndata=0;
+ int nfiles=0;
  index_t count_out=0,count_in=0,total=0;
  size_t datasize,item;
  FILESORT_VARINT nextitemsize,largestitemsize=0;
- char *data;
- void **datap;
  thread_data *threads;
- int i,more=1;
+ int nthreads,i,more=1;
  char *filename=(char*)malloc_logassert(strlen(option_tmpdirname)+24);
 #if defined(USE_PTHREADS) && USE_PTHREADS
- int nthreads=0;
+ int threads_running=0;
 #endif
 
  /* Allocate the RAM buffer and other bits */
@@ -542,20 +394,33 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
 
  /* We can not know in advance how many data items there are.  Each
     one will require RAM for data, FILESORT_VARALIGN and sizeof(void*)
-    Assume that data+FILESORT_VARALIGN+sizeof(void*) is 4*data. */
+    Assume that data+FILESORT_VARALIGN+sizeof(void*) is 2*data (tests
+    show values of between 1.2 and 2.5). */
 
- if((datasize*4)<option_filesort_ramsize/option_filesort_threads)
-    /* use one thread */;
- else if((datasize*4)<option_filesort_ramsize)
-    datasize=(datasize*4)/option_filesort_threads;
+ datasize=datasize*2;
+
+ if(option_filesort_threads==1)
+    nthreads = 1;
+ else if(datasize<option_filesort_ramsize)
+    nthreads = 1 + datasize / (option_filesort_ramsize/option_filesort_threads);
  else
-    datasize=option_filesort_ramsize/option_filesort_threads;
+    nthreads = option_filesort_threads;
+
+ if(datasize>option_filesort_ramsize)
+    datasize=option_filesort_ramsize;
+
+ datasize=datasize/nthreads;
 
  datasize=FILESORT_VARALIGN*((datasize+FILESORT_VARALIGN-1)/FILESORT_VARALIGN);
 
- threads=(thread_data*)calloc_logassert(option_filesort_threads,sizeof(thread_data));
+#if DEBUG
+ printf("DEBUG: filesort_vary datasize=%lu option_ramsize/option_threads=%lu => datasize=%lu nthreads=%d\n",
+        2*(size_t)SizeFileFD(fd_in),option_filesort_ramsize/option_filesort_threads,datasize,nthreads);
+#endif
 
- for(i=0;i<option_filesort_threads;i++)
+ threads=(thread_data*)calloc_logassert(nthreads,sizeof(thread_data));
+
+ for(i=0;i<nthreads;i++)
    {
     threads[i].fd=-1;
 
@@ -564,7 +429,12 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
 
     log_malloc(threads[i].data,datasize);
 
+    threads[i].itemsize=0;
     threads[i].compare=compare_function;
+
+#if defined(USE_PTHREADS) && USE_PTHREADS
+    threads[i].running=0;
+#endif
    }
 
  /* Loop around, fill the buffer, sort the data and write a temporary file */
@@ -579,48 +449,16 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
-    if(option_filesort_threads>1)
-      {
-       /* If all threads are in use wait for an existing thread to finish */
+    /* Choose a thread to use */
 
-       if(nthreads==option_filesort_threads)
-         {
-          pthread_mutex_lock(&running_mutex);
-
-          while(nthreads==option_filesort_threads)
-            {
-             for(i=0;i<option_filesort_threads;i++)
-                if(threads[i].running==2)
-                  {
-                   pthread_join(threads[i].thread,NULL);
-                   threads[i].running=0;
-                   CloseFileBuffered(threads[i].fd);
-                   nthreads--;
-                  }
-
-             if(nthreads==option_filesort_threads)
-                pthread_cond_wait(&running_cond,&running_mutex);
-            }
-
-          pthread_mutex_unlock(&running_mutex);
-         }
-
-       /* Find a spare slot */
-
-       pthread_mutex_lock(&running_mutex);
-
-       for(thread=0;thread<option_filesort_threads;thread++)
-          if(!threads[thread].running)
-             break;
-
-       pthread_mutex_unlock(&running_mutex);
-      }
+    if(nthreads>1)
+       thread=wait_for_free_thread(threads,nthreads,&threads_running);
 
 #endif
 
     threads[thread].datap=(void**)(threads[thread].data+datasize);
 
-    threads[thread].n=0;
+    item=0;
 
     /* Read in the data and create pointers */
 
@@ -647,7 +485,7 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
           ramused+=FILESORT_VARALIGN-FILESORT_VARSIZE;
 
           total++;
-          threads[thread].n++;
+          item++;
          }
        else
           ramused-=FILESORT_VARSIZE;
@@ -661,16 +499,27 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
          }
       }
 
+#if DEBUG
+    printf("DEBUG: filesort_vary thread=%d file=%d item=%lu total=%u ramused=%lu more=%d\n",thread,nfiles,item,total,ramused+item*sizeof(void*),more);
+#endif
+
     /* No new data read in this time round */
 
-    if(threads[thread].n==0)
+    if(item==0)
        break;
 
+    /* Update the number of items to sort */
+
+    threads[thread].n=item;
+
     /* Shortcut if only one file, don't write to disk */
 
     if(more==0 && nfiles==0)
-       filesort_heapsort(threads[thread].datap,threads[thread].n,threads[thread].compare);
+      {
+       filename[0]=0;
 
+       filesort_heapsort_thread(&threads[thread]);
+      }
     else
       {
        /* Create the file descriptor (not thread-safe) */
@@ -679,9 +528,9 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
 
        threads[thread].fd=OpenFileBufferedNew(filename);
 
-       if(option_filesort_threads==1)
+       if(nthreads==1)
          {
-          filesort_vary_heapsort_thread(&threads[thread]);
+          filesort_heapsort_thread(&threads[thread]);
 
           CloseFileBuffered(threads[thread].fd);
          }
@@ -690,15 +539,11 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
 
        else
          {
-          pthread_mutex_lock(&running_mutex);
-
           threads[thread].running=1;
 
-          pthread_mutex_unlock(&running_mutex);
+          pthread_create(&threads[thread].thread,NULL,(void* (*)(void*))filesort_heapsort_thread,&threads[thread]);
 
-          pthread_create(&threads[thread].thread,NULL,(void* (*)(void*))filesort_vary_heapsort_thread,&threads[thread]);
-
-          nthreads++;
+          threads_running++;
          }
 
 #endif
@@ -709,27 +554,12 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
    }
  while(more);
 
- /* Wait for all of the threads to finish */
-
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
- while(option_filesort_threads>1 && nthreads)
-   {
-    pthread_mutex_lock(&running_mutex);
+ /* Wait for all of the threads to finish */
 
-    pthread_cond_wait(&running_cond,&running_mutex);
-
-    for(i=0;i<option_filesort_threads;i++)
-       if(threads[i].running==2)
-         {
-          pthread_join(threads[i].thread,NULL);
-          threads[i].running=0;
-          CloseFileBuffered(threads[i].fd);
-          nthreads--;
-         }
-
-    pthread_mutex_unlock(&running_mutex);
-   }
+ if(nthreads>1)
+    wait_for_all_threads(threads,nthreads,&threads_running);
 
 #endif
 
@@ -753,161 +583,23 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
          }
       }
 
-    DeleteFile(filename);
+    if(filename[0])
+       DeleteFile(filename);
 
     goto tidy_and_exit;
    }
 
- /* Check that number of files is less than file size */
+ /* Open all of the temporary files and perform an n-way merge using a binary heap */
 
  largestitemsize=FILESORT_VARALIGN*((largestitemsize+FILESORT_VARALIGN-1)/FILESORT_VARALIGN);
 
- logassert((unsigned)nfiles<((datasize-nfiles*sizeof(void*))/(FILESORT_VARALIGN+largestitemsize)),"Too many temporary files (use more sorting memory?)");
-
- /* Open all of the temporary files */
-
- fds=(int*)malloc_logassert(nfiles*sizeof(int));
-
- for(i=0;i<nfiles;i++)
-   {
-    sprintf(filename,"%s/filesort.%d.tmp",option_tmpdirname,i);
-
-    fds[i]=ReOpenFileBuffered(filename);
-
-    DeleteFile(filename);
-   }
-
- /* Perform an n-way merge using a binary heap */
-
- heap=(int*)malloc_logassert((1+nfiles)*sizeof(int));
-
- data=threads[0].data;
- datap=(void**)(data+datasize-nfiles*sizeof(void*));
-
- /* Fill the heap to start with */
-
- for(i=0;i<nfiles;i++)
-   {
-    int index;
-    FILESORT_VARINT itemsize;
-
-    datap[i]=data+FILESORT_VARALIGN+i*(largestitemsize+FILESORT_VARALIGN);
-
-    ReadFileBuffered(fds[i],&itemsize,FILESORT_VARSIZE);
-
-    *(FILESORT_VARINT*)((char*)datap[i]-FILESORT_VARSIZE)=itemsize;
-
-    ReadFileBuffered(fds[i],datap[i],itemsize);
-
-    index=i+1;
-
-    heap[index]=i;
-
-    /* Bubble up the new value */
-
-    while(index>1)
-      {
-       int newindex;
-       int temp;
-
-       newindex=index/2;
-
-       if(compare_function(datap[heap[index]],datap[heap[newindex]])>=0)
-          break;
-
-       temp=heap[index];
-       heap[index]=heap[newindex];
-       heap[newindex]=temp;
-
-       index=newindex;
-      }
-   }
-
- /* Repeatedly pull out the root of the heap and refill from the same file */
-
- ndata=nfiles;
-
- do
-   {
-    int index=1;
-    FILESORT_VARINT itemsize;
-
-    if(!post_sort_function || post_sort_function(datap[heap[index]],count_out))
-      {
-       itemsize=*(FILESORT_VARINT*)((char*)datap[heap[index]]-FILESORT_VARSIZE);
-
-       WriteFileBuffered(fd_out,(char*)datap[heap[index]]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
-       count_out++;
-      }
-
-    if(ReadFileBuffered(fds[heap[index]],&itemsize,FILESORT_VARSIZE))
-      {
-       heap[index]=heap[ndata];
-       ndata--;
-      }
-    else
-      {
-       *(FILESORT_VARINT*)((char*)datap[heap[index]]-FILESORT_VARSIZE)=itemsize;
-
-       ReadFileBuffered(fds[heap[index]],datap[heap[index]],itemsize);
-      }
-
-    /* Bubble down the new value */
-
-    while((2*index)<ndata)
-      {
-       int newindex;
-       int temp;
-
-       newindex=2*index;
-
-       if(compare_function(datap[heap[newindex]],datap[heap[newindex+1]])>=0)
-          newindex=newindex+1;
-
-       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
-          break;
-
-       temp=heap[newindex];
-       heap[newindex]=heap[index];
-       heap[index]=temp;
-
-       index=newindex;
-      }
-
-    if((2*index)==ndata)
-      {
-       int newindex;
-       int temp;
-
-       newindex=2*index;
-
-       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
-          ; /* break */
-       else
-         {
-          temp=heap[newindex];
-          heap[newindex]=heap[index];
-          heap[index]=temp;
-         }
-      }
-   }
- while(ndata>0);
+ count_out=filesort_merge(fd_out,nfiles,0,largestitemsize,compare_function,post_sort_function);
 
  /* Tidy up */
 
  tidy_and_exit:
 
- if(fds)
-   {
-    for(i=0;i<nfiles;i++)
-       CloseFileBuffered(fds[i]);
-    free(fds);
-   }
-
- if(heap)
-    free(heap);
-
- for(i=0;i<option_filesort_threads;i++)
+ for(i=0;i<nthreads;i++)
    {
     log_free(threads[i].data);
 
@@ -922,56 +614,111 @@ index_t filesort_vary(int fd_in,int fd_out,int (*pre_sort_function)(void*,index_
 }
 
 
-/*++++++++++++++++++++++++++++++++++++++
-  A wrapper function that can be run in a thread for fixed data.
-
-  void *filesort_fixed_heapsort_thread Returns NULL (required to return void*).
-
-  thread_data *thread The data to be processed in this thread.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-static void *filesort_fixed_heapsort_thread(thread_data *thread)
-{
- size_t item;
-
- /* Sort the data pointers using a heap sort */
-
- filesort_heapsort(thread->datap,thread->n,thread->compare);
-
- /* Write the result to the given temporary file */
-
- if(thread->fd > 0)
-    for(item=0;item<thread->n;item++)
-       WriteFileBuffered(thread->fd,thread->datap[item],thread->itemsize);
-
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
- if(option_filesort_threads>1)
+/*++++++++++++++++++++++++++++++++++++++
+  Wait for a thread to become free and return the number of that thread while updating the number running.
+
+  int wait_for_free_thread Returns the thread number.
+
+  thread_data *threads The thread data.
+
+  int nthreads The number of threads allowed.
+
+  int *threads_running The number of threads, updated when called.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int wait_for_free_thread(thread_data *threads,int nthreads,int *threads_running)
+{
+ int i;
+
+ /* Lock the mutex */
+
+ pthread_mutex_lock(&running_mutex);
+
+ /* If all threads are in use wait for an existing thread to finish */
+
+ while((*threads_running)==nthreads)
    {
-    pthread_mutex_lock(&running_mutex);
+    for(i=0;i<nthreads;i++)
+       if(threads[i].running==2)
+         {
+          pthread_join(threads[i].thread,NULL);
+          threads[i].running=0;
+          CloseFileBuffered(threads[i].fd);
+          (*threads_running)--;
+         }
 
-    thread->running=2;
-
-    pthread_cond_signal(&running_cond);
-
-    pthread_mutex_unlock(&running_mutex);
+    if((*threads_running)==nthreads)
+       pthread_cond_wait(&running_cond,&running_mutex);
    }
 
-#endif
+ /* Find a spare slot */
 
- return(NULL);
+ for(i=0;i<nthreads;i++)
+    if(!threads[i].running)
+       break;
+
+ /* Unlock the mutex */
+
+ pthread_mutex_unlock(&running_mutex);
+
+ return(i);
 }
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  A wrapper function that can be run in a thread for variable data.
+  Wait for all threads to finish while updating the number running.
 
-  void *filesort_vary_heapsort_thread Returns NULL (required to return void*).
+  thread_data *threads The thread data.
+
+  int nthreads The number of threads allowed.
+
+  int *threads_running The number of threads, updated when called.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void wait_for_all_threads(thread_data *threads,int nthreads,int *threads_running)
+{
+ int i;
+
+ /* Lock the mutex */
+
+ pthread_mutex_lock(&running_mutex);
+
+ /* Wait for all threads to finish */
+
+ while(*threads_running)
+   {
+    for(i=0;i<nthreads;i++)
+       if(threads[i].running==2)
+         {
+          pthread_join(threads[i].thread,NULL);
+          threads[i].running=0;
+          CloseFileBuffered(threads[i].fd);
+          (*threads_running)--;
+         }
+
+    if(*threads_running)
+       pthread_cond_wait(&running_cond,&running_mutex);
+   }
+
+ /* Unlock the mutex */
+
+ pthread_mutex_unlock(&running_mutex);
+}
+
+#endif
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  A wrapper function that can be run in a thread for fixed or variable data.
+
+  void *filesort_heapsort_thread Returns NULL (required to return void*).
 
   thread_data *thread The data to be processed in this thread.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void *filesort_vary_heapsort_thread(thread_data *thread)
+static void *filesort_heapsort_thread(thread_data *thread)
 {
  size_t item;
 
@@ -979,19 +726,27 @@ static void *filesort_vary_heapsort_thread(thread_data *thread)
 
  filesort_heapsort(thread->datap,thread->n,thread->compare);
 
- /* Write the result to the given temporary file */
+ /* Write the result to the temporary file if given */
 
- if(thread->fd > 0)
-    for(item=0;item<thread->n;item++)
-      {
-       FILESORT_VARINT itemsize=*(FILESORT_VARINT*)((char*)thread->datap[item]-FILESORT_VARSIZE);
+ if(thread->fd>0)
+   {
+    if(thread->itemsize>0)
+       for(item=0;item<thread->n;item++)
+          WriteFileBuffered(thread->fd,thread->datap[item],thread->itemsize);
+    else
+       for(item=0;item<thread->n;item++)
+         {
+          FILESORT_VARINT itemsize=*(FILESORT_VARINT*)((char*)thread->datap[item]-FILESORT_VARSIZE);
 
-       WriteFileBuffered(thread->fd,(char*)thread->datap[item]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
-      }
+          WriteFileBuffered(thread->fd,(char*)thread->datap[item]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
+         }
+   }
 
 #if defined(USE_PTHREADS) && USE_PTHREADS
 
- if(option_filesort_threads>1)
+ /* Signal that this thread has finished */
+
+ if(thread->running==1)
    {
     pthread_mutex_lock(&running_mutex);
 
@@ -1105,3 +860,219 @@ void filesort_heapsort(void **datap,size_t nitems,int(*compare_function)(const v
       }
    }
 }
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  A function to merge an array of sorted files efficiently.
+
+  The data is merged using an "external sort" http://en.wikipedia.org/wiki/External_sorting
+  where only one item is read from each file at a time.
+
+  index_t filesort_merge Returns the number of items written to the output file
+
+  int fd_out The file descriptor of the output file (opened for writing and empty).
+
+  int nfiles The number of files to open and merge
+
+  size_t itemsize The size of each item (non-zero if a fixed size sort).
+
+  size_t largestitemsize The maximum size of each item (non-zero if a variable size sort).
+
+  int (*compare_function)(const void*, const void*) The comparison function.  This is identical
+     to qsort if the data to be sorted is an array of things not pointers.
+
+  int (*post_sort_function)(void *,index_t) If non-NULL then this function is called for
+     each item after they have been sorted.  The second parameter is the number of objects
+     already written to the output file.  If the function returns 1 then the object is written
+     to the output file., otherwise it is ignored.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static index_t filesort_merge(int fd_out,int nfiles,size_t itemsize,size_t largestitemsize,
+                              int(*compare_function)(const void*, const void*),
+                              int (*post_sort_function)(void*,index_t))
+{
+ int ndata=0;
+ char *data,*filename;
+ void **datap;
+ int *fds,*heap;
+ index_t count_out=0;
+ int i;
+
+ /* Allocate the memory */
+
+ filename=(char*)malloc_logassert(strlen(option_tmpdirname)+24);
+
+ heap=(int*)malloc_logassert((1+nfiles)*sizeof(int));
+
+ if(itemsize)
+    data=(void*)malloc_logassert(nfiles*itemsize);
+ else
+    data=(void*)malloc_logassert(nfiles*(largestitemsize+FILESORT_VARALIGN));
+
+ datap=(void**)malloc_logassert(nfiles*sizeof(void*));
+
+ fds=(int*)malloc_logassert(nfiles*sizeof(int));
+
+ /* Re-open the files */
+
+ for(i=0;i<nfiles;i++)
+   {
+    sprintf(filename,"%s/filesort.%d.tmp",option_tmpdirname,i);
+
+    fds[i]=ReOpenFileBuffered(filename);
+
+    DeleteFile(filename);
+   }
+
+ /* Fill the heap to start with */
+
+ for(i=0;i<nfiles;i++)
+   {
+    int index;
+
+    if(itemsize)
+      {
+       datap[i]=data+i*itemsize;
+
+       ReadFileBuffered(fds[i],datap[i],itemsize);
+      }
+    else
+      {
+       FILESORT_VARINT itemsize;
+
+       datap[i]=data+FILESORT_VARALIGN+i*(largestitemsize+FILESORT_VARALIGN);
+
+       ReadFileBuffered(fds[i],&itemsize,FILESORT_VARSIZE);
+
+       *(FILESORT_VARINT*)((char*)datap[i]-FILESORT_VARSIZE)=itemsize;
+
+       ReadFileBuffered(fds[i],datap[i],itemsize);
+      }
+
+    index=i+1;
+
+    heap[index]=i;
+
+    /* Bubble up the new value */
+
+    while(index>1)
+      {
+       int newindex;
+       int temp;
+
+       newindex=index/2;
+
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])>=0)
+          break;
+
+       temp=heap[index];
+       heap[index]=heap[newindex];
+       heap[newindex]=temp;
+
+       index=newindex;
+      }
+   }
+
+ /* Repeatedly pull out the root of the heap and refill from the same file */
+
+ ndata=nfiles;
+
+ do
+   {
+    int index=1;
+
+    if(!post_sort_function || post_sort_function(datap[heap[index]],count_out))
+      {
+       if(itemsize)
+          WriteFileBuffered(fd_out,datap[heap[index]],itemsize);
+       else
+         {
+          FILESORT_VARINT itemsize=*(FILESORT_VARINT*)((char*)datap[heap[index]]-FILESORT_VARSIZE);
+
+          WriteFileBuffered(fd_out,(char*)datap[heap[index]]-FILESORT_VARSIZE,itemsize+FILESORT_VARSIZE);
+         }
+       count_out++;
+      }
+
+    if(itemsize)
+      {
+       if(ReadFileBuffered(fds[heap[index]],datap[heap[index]],itemsize))
+         {
+          heap[index]=heap[ndata];
+          ndata--;
+         }
+      }
+    else
+      {
+       FILESORT_VARINT itemsize;
+
+       if(ReadFileBuffered(fds[heap[index]],&itemsize,FILESORT_VARSIZE))
+         {
+          heap[index]=heap[ndata];
+          ndata--;
+         }
+       else
+         {
+          *(FILESORT_VARINT*)((char*)datap[heap[index]]-FILESORT_VARSIZE)=itemsize;
+
+          ReadFileBuffered(fds[heap[index]],datap[heap[index]],itemsize);
+         }
+      }
+
+    /* Bubble down the new value */
+
+    while((2*index)<ndata)
+      {
+       int newindex;
+       int temp;
+
+       newindex=2*index;
+
+       if(compare_function(datap[heap[newindex]],datap[heap[newindex+1]])>=0)
+          newindex=newindex+1;
+
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
+          break;
+
+       temp=heap[newindex];
+       heap[newindex]=heap[index];
+       heap[index]=temp;
+
+       index=newindex;
+      }
+
+    if((2*index)==ndata)
+      {
+       int newindex;
+       int temp;
+
+       newindex=2*index;
+
+       if(compare_function(datap[heap[index]],datap[heap[newindex]])<=0)
+          ; /* break */
+       else
+         {
+          temp=heap[newindex];
+          heap[newindex]=heap[index];
+          heap[index]=temp;
+         }
+      }
+   }
+ while(ndata>0);
+
+ /* Tidy up */
+
+ free(filename);
+
+ for(i=0;i<nfiles;i++)
+    CloseFileBuffered(fds[i]);
+
+ free(fds);
+
+ free(heap);
+
+ free(data);
+ free(datap);
+
+ return(count_out);
+}
