@@ -3,7 +3,7 @@
 
  Part of the Routino routing software.
  ******************/ /******************
- This file Copyright 2008-2015, 2018, 2019 Andrew M. Bishop
+ This file Copyright 2008-2015, 2018, 2019, 2020, 2022, 2024 Andrew M. Bishop
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as published by
@@ -113,6 +113,11 @@ WaysX *NewWayList(int append,int readonly)
  log_malloc(waysx->cache,sizeof(*waysx->cache));
 #endif
 
+ waysx->ifilename_tmp=(char*)malloc_logassert(strlen(option_tmpdirname)+40); /* allow %p to be up to 20 bytes */
+ waysx->ofilename_tmp=(char*)malloc_logassert(strlen(option_tmpdirname)+40); /* allow %p to be up to 20 bytes */
+
+ sprintf(waysx->ifilename_tmp,"%s/waysx.%p.idx.tmp",option_tmpdirname,(void*)waysx);
+ sprintf(waysx->ofilename_tmp,"%s/waysx.%p.off.tmp",option_tmpdirname,(void*)waysx);
 
  waysx->nfilename_tmp=(char*)malloc_logassert(strlen(option_tmpdirname)+40); /* allow %p to be up to 20 bytes */
 
@@ -140,17 +145,11 @@ void FreeWayList(WaysX *waysx,int keep)
  free(waysx->filename);
  free(waysx->filename_tmp);
 
- if(waysx->idata)
-   {
-    log_free(waysx->idata);
-    free(waysx->idata);
-   }
+ DeleteFile(waysx->ifilename_tmp);
 
- if(waysx->odata)
-   {
-    log_free(waysx->odata);
-    free(waysx->odata);
-   }
+ DeleteFile(waysx->ofilename_tmp);
+
+ free(waysx->ifilename_tmp);
 
  if(waysx->cdata)
    {
@@ -189,7 +188,7 @@ void FreeWayList(WaysX *waysx,int keep)
 
 void AppendWayList(WaysX *waysx,way_t id,Way *way,node_t *nodes,int nnodes,const char *name)
 {
- WayX wayx;
+ WayX wayx={0};
  uint64_t longsize;
  FILESORT_VARINT size;
  node_t nonode=NO_NODE_ID;
@@ -199,7 +198,16 @@ void AppendWayList(WaysX *waysx,way_t id,Way *way,node_t *nodes,int nnodes,const
 
  longsize=sizeof(WayX)+(nnodes+1)*sizeof(node_t)+strlen(name)+1;
 
- logassert(longsize<FILESORT_MAXINT,"Way contains too much data (change FILESORT_VARINT to 32-bits?)"); /* Ensure no overflow of FILESORT_VARINT integer */
+ if(longsize>=FILESORT_MAXINT) /* Ensure no overflow of FILESORT_VARINT integer */
+   {
+    logerror("Way %"Pway_t" contains too much data; ignoring some nodes (or change FILESORT_VARINT to 32-bits?)\n",logerror_way(id));
+
+    nnodes=(FILESORT_MAXINT-(sizeof(WayX)+strlen(name)+1))/sizeof(node_t)-1;
+
+    longsize=sizeof(WayX)+(nnodes+1)*sizeof(node_t)+strlen(name)+1;
+
+    logassert(longsize<FILESORT_MAXINT,"Way still contains too much data (change FILESORT_VARINT to 32-bits?)\n");
+   }
 
  size=longsize;
 
@@ -249,35 +257,28 @@ index_t IndexWayX(WaysX *waysx,way_t id)
  if(waysx->number==0)           /* There are no ways */
     return(NO_WAY);
 
- if(id<waysx->idata[start])     /* Key is before start */
-    return(NO_WAY);
-
- if(id>waysx->idata[end])       /* Key is after end */
-    return(NO_WAY);
-
  /* Binary search - search key exact match only is required.
   *
-  *  # <- start  |  Check mid and move start or end if it doesn't match
+  *  # <- start  |  Check mid and exit if it matches else move start or end.
   *  #           |
   *  #           |  Since an exact match is wanted we can set end=mid-1
-  *  # <- mid    |  or start=mid+1 because we know that mid doesn't match.
+  *  # <- mid    |  or start=mid+1 if we find that mid doesn't match.
   *  #           |
   *  #           |  Eventually either end=start or end=start+1 and one of
-  *  # <- end    |  start or end is the wanted one.
+  *  # <- end    |  start or end is the wanted one or neither is.
   */
 
- do
+ while((end-start)>1)
    {
     mid=start+(end-start)/2;      /* Choose mid point (avoid overflow) */
 
     if(waysx->idata[mid]<id)      /* Mid point is too low */
        start=mid+1;
     else if(waysx->idata[mid]>id) /* Mid point is too high */
-       end=mid?(mid-1):mid;
+       end=mid-1;
     else                          /* Mid point is correct */
        return(mid);
    }
- while((end-start)>1);
 
  if(waysx->idata[start]==id)      /* Start is correct */
     return(start);
@@ -308,10 +309,9 @@ void SortWayList(WaysX *waysx)
 
  fd=ReplaceFileBuffered(waysx->filename_tmp,&waysx->fd);
 
- /* Allocate the array of indexes */
+ /* Open a file for the index */
 
- waysx->idata=(way_t*)malloc_logassert(waysx->number*sizeof(way_t));
- log_malloc(waysx->idata,waysx->number*sizeof(way_t));
+ waysx->ifd=OpenFileBufferedNew(waysx->ifilename_tmp);
 
  /* Sort the ways by ID and index them */
 
@@ -330,6 +330,8 @@ void SortWayList(WaysX *waysx)
  waysx->fd=CloseFileBuffered(waysx->fd);
  CloseFileBuffered(fd);
 
+ waysx->ifd=CloseFileBuffered(waysx->ifd);
+
  /* Print the final message */
 
  printf_last("Sorted Ways: Ways=%"Pindex_t" Duplicates=%"Pindex_t,xnumber,xnumber-waysx->number);
@@ -382,7 +384,7 @@ static int deduplicate_and_index_by_id(WayX *wayx,index_t index)
        return(0);
     else
       {
-       sortwaysx->idata[index]=wayx->id;
+       WriteFileBuffered(sortwaysx->ifd,&wayx->id,sizeof(way_t));
 
        return(1);
       }
@@ -433,6 +435,11 @@ SegmentsX *SplitWays(WaysX *waysx,NodesX *nodesx,int keep)
 
  nfd=OpenFileBufferedNew(waysx->nfilename_tmp);
 
+ /* Map the index into memory */
+
+ nodesx->idata=MapFile(nodesx->ifilename_tmp);
+ waysx->idata =MapFile(waysx->ifilename_tmp);
+
  /* Loop through the ways and create the segments and way names */
 
  for(i=0;i<waysx->number;i++)
@@ -511,6 +518,11 @@ SegmentsX *SplitWays(WaysX *waysx,NodesX *nodesx,int keep)
 
  CloseFileBuffered(nfd);
 
+ /* Unmap the index from memory */
+
+ nodesx->idata=UnmapFile(nodesx->idata);
+ waysx->idata =UnmapFile(waysx->idata);
+
  /* Print the final message */
 
  printf_last("Split Ways: Ways=%"Pindex_t" Segments=%"Pindex_t,waysx->number,segmentsx->number);
@@ -528,11 +540,10 @@ SegmentsX *SplitWays(WaysX *waysx,NodesX *nodesx,int keep)
 
 void SortWayNames(WaysX *waysx)
 {
- index_t i;
+ index_t i,nnames=0;
  int nfd;
  char *names[2]={NULL,NULL};
  uint32_t namelen[2]={0,0};
- int nnames=0;
  uint32_t lastlength=0;
 
  /* Print the start message */
@@ -565,7 +576,7 @@ void SortWayNames(WaysX *waysx)
 
  printf_first("Updating Ways with Names: Ways=0 Names=0");
 
- /* Map into memory /  open the file */
+ /* Map into memory / open the file */
 
 #if !SLIM
  waysx->data=MapFileWriteable(waysx->filename_tmp);
@@ -793,8 +804,6 @@ static int deduplicate_and_index_by_compact_id(WayX *wayx,index_t index)
 
     sortwaysx->cdata[wayx->id]=index;
 
-    wayx->id=index;
-
     return(1);
    }
  else
